from typing import Callable, Sequence, List, Tuple

import numpy as np
from torch import Tensor

from treeboost_autograd.pytorch_objective import PytorchObjective


class SklearnStyleObjective(PytorchObjective):
    def __init__(self, loss_function: Callable[[Tensor, Tensor], Tensor]):
        super().__init__(loss_function, use_minus_loss_as_objective=False)

    def __call__(self, targets: np.ndarray, preds: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        return self.calculate_derivatives(preds, targets)


class LightGbmObjective(SklearnStyleObjective):
    pass


class XgboostObjective(SklearnStyleObjective):
    pass


class CatboostObjective(PytorchObjective):
    def __init__(self, loss_function: Callable[[Tensor, Tensor], Tensor]):
        super().__init__(loss_function, use_minus_loss_as_objective=True)

    def calc_ders_range(self,
                        preds: Sequence[float],
                        targets: Sequence[float],
                        weights: Sequence[float] = None
                        ) -> List[Tuple[float, float]]:
        deriv1, deriv2 = self.calculate_derivatives(preds, targets, weights)
        result = list(zip(deriv1, deriv2))
        return result



from typing import List, Tuple, Union, Callable

import numpy as np
import numpy.random as npr
import pytest
from torch import Tensor
from torch.nn import BCEWithLogitsLoss, MSELoss

from tests.official_catboost_objectives import OfficialLoglossObjective, OfficialMseObjective
from treeboost_autograd.booster_objectives import CatboostObjective


class _TestParams:
    @classmethod
    def generate_test_params(cls) -> List[Tuple]:
        classification_params = [
            (approxes, targets, weights, OfficialLoglossObjective(), BCEWithLogitsLoss(reduction="sum"))
            for approxes, targets, weights in cls._generate_approxes_targets_weights(targets_type="classification")
        ]
        regression_params = [
            (approxes, targets, weights, OfficialMseObjective(), cls._mse_loss)
            for approxes, targets, weights in cls._generate_approxes_targets_weights(targets_type="regression")
        ]
        test_params = classification_params + regression_params
        return test_params

    @staticmethod
    def _generate_approxes_targets_weights(
            targets_type: str,
            n_different: int = 10,
            vector_size: int = 100,
            random_seed: int = 34
    ) -> List[Tuple[np.ndarray, np.ndarray, Union[np.ndarray, None]]]:
        assert targets_type in ("classification", "regression")

        random_state = npr.RandomState(random_seed)
        test_tuples = []
        for _ in range(n_different):
            approxes = random_state.randn(vector_size)
            if targets_type == "classification":
                targets = random_state.randint(low=0, high=2, size=vector_size)
            elif targets_type == "regression":
                targets = random_state.randn(vector_size)
            weights = random_state.rand(vector_size)
            test_tuples.append((approxes, targets, weights))
            test_tuples.append((approxes, targets, None))
        return test_tuples

    @staticmethod
    def _mse_loss(approxes: Tensor, targets: Tensor) -> Tensor:
        return 0.5 * MSELoss(reduction="sum")(approxes, targets)


class TestCatboostObjective:
    @pytest.mark.parametrize("approxes,targets,weights,official_objective,pytorch_loss_func",
                             _TestParams.generate_test_params())
    def test_calc_ders_range(self,
                             approxes: np.ndarray,
                             targets: np.ndarray,
                             weights: Union[np.ndarray, None],
                             official_objective: Union[OfficialLoglossObjective, OfficialMseObjective],
                             pytorch_loss_func: Callable[[Tensor, Tensor], Tensor]):
        custom_pytorch_objective = CatboostObjective(pytorch_loss_func)
        result_pytorch = custom_pytorch_objective.calc_ders_range(approxes, targets, weights)
        result_official = official_objective.calc_ders_range(approxes, targets, weights)
        assert np.allclose(result_official, result_pytorch, rtol=1e-4)



from itertools import product
from typing import Callable

import pytest

from examples.example_classification import train_and_eval_custom_classifier, soft_margin_loss, squared_hinge_loss


class TestTrainAndEvalCustomClassifier:
    @pytest.mark.parametrize("boosting_package,custom_loss_function",
                             product(["catboost", "xgboost", "lightgbm"], [soft_margin_loss, squared_hinge_loss]))
    def test_train_and_eval_custom_classifier(self, boosting_package: str, custom_loss_function: Callable):
        macro_f1_score = train_and_eval_custom_classifier(boosting_package, custom_loss_function)
        assert macro_f1_score > 0.9



import pytest

from examples.example_regression import train_and_eval_custom_regressor, dont_undershoot_loss


class TestTrainAndEvalCustomClassifier:
    @pytest.mark.parametrize("boosting_package,n_estimators",
                             [("xgboost", 100), ("lightgbm", 100), ("catboost", 300)])
    def test_train_and_eval_custom_classifier(self, boosting_package: str, n_estimators: int):
        mean_abs_rel_diff, undershoot_percentage = train_and_eval_custom_regressor(
            boosting_package, dont_undershoot_loss, n_estimators, is_plot=False)
        assert mean_abs_rel_diff < 0.3
        assert undershoot_percentage < 0.3



from typing import Callable, Sequence, Tuple

import numpy as np
import torch
from torch import Tensor


class PytorchObjective:
    def __init__(self,
                 loss_function: Callable[[Tensor, Tensor], Tensor],
                 use_minus_loss_as_objective: bool):
        self.loss_function = loss_function
        self.sign = -1 if use_minus_loss_as_objective else 1

    def calculate_derivatives(self,
                              preds: Sequence[float],
                              targets: Sequence[float],
                              weights: Sequence[float] = None
                              ) -> Tuple[np.ndarray, np.ndarray]:
        """  Calculates first and second derivatives of the objective w.r.t preds.  """
        preds = torch.FloatTensor(preds).requires_grad_()
        targets = torch.FloatTensor(targets)

        objective = self.sign * self.loss_function(preds, targets)
        deriv1, deriv2 = self._calculate_derivatives(objective, preds)

        if weights is not None:
            weights = np.asarray(weights)
            deriv1, deriv2 = weights * deriv1, weights * deriv2

        return deriv1, deriv2

    @staticmethod
    def _calculate_derivatives(objective: Tensor, preds: Tensor) -> Tuple[np.ndarray, np.ndarray]:
        deriv1, = torch.autograd.grad(objective, preds, create_graph=True)

        deriv2 = []
        for i in range(len(preds)):
            deriv2_i, = torch.autograd.grad(deriv1[i], preds, retain_graph=True)
            deriv2_i = deriv2_i[i].item()
            deriv2.append(deriv2_i)

        deriv1 = deriv1.detach().numpy()
        deriv2 = np.array(deriv2)
        return deriv1, deriv2



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['treeboost_autograd']

package_data = \
{'': ['*']}

install_requires = \
['numpy', 'torch']

extras_require = \
{'dev': ['catboost',
         'xgboost',
         'lightgbm',
         'matplotlib',
         'pytest',
         'scikit_learn']}

setup(name='treeboost_autograd',
      version='0.1.1',
      description='treeboost_autograd',
      author='Tomer Ronen',
      author_email='tomer.ronen34@gmail.com',
      url='https://github.com/TomerRonen34/treeboost_autograd',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )



from typing import Union, Callable, Tuple

import matplotlib.pyplot as plt
import numpy as np
import torch
from sklearn.datasets import load_boston
from sklearn.model_selection import train_test_split
from torch import Tensor

from catboost import CatBoostRegressor
from lightgbm import LGBMRegressor
from xgboost import XGBRegressor

from treeboost_autograd import CatboostObjective, LightGbmObjective, XgboostObjective


def main():
    _plot_custom_loss(dont_undershoot_loss)
    train_and_eval_custom_regressor("xgboost", dont_undershoot_loss, n_estimators=100)
    train_and_eval_custom_regressor("lightgbm", dont_undershoot_loss, n_estimators=100)
    train_and_eval_custom_regressor("catboost", dont_undershoot_loss, n_estimators=300)


def dont_undershoot_loss(preds: Tensor, targets: Tensor) -> Tensor:
    """ penalize undershooting much more than overshooting (regression loss) """
    raw_diff = preds - targets
    is_undershoot = raw_diff < 0
    loss = torch.zeros_like(raw_diff)
    loss[is_undershoot] = torch.abs(raw_diff[is_undershoot] - 1) ** 4 - 1
    loss[~is_undershoot] = torch.abs(raw_diff[~is_undershoot])
    loss = loss.sum()
    return loss


def train_and_eval_custom_regressor(boosting_package: str, custom_loss_function: Callable[[Tensor, Tensor], Tensor],
                                    n_estimators: int, random_seed: int = 2021, is_plot: bool = True
                                    ) -> Tuple[float, float]:
    assert boosting_package in ["catboost", "xgboost", "lightgbm"]
    _print_title(boosting_package)

    X, y = load_boston(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=2021)

    model = _fit_custom_regressor(boosting_package,
                                  X_train, y_train, custom_loss_function,
                                  n_estimators, random_seed)

    pred_test = model.predict(X_test)
    mean_abs_rel_diff = _calc_mean_absolute_relative_difference(pred_test, y_test)
    undershoot_percentage = _calc_undershoot_percentage(pred_test, y_test)

    if is_plot:
        plot_title = _generate_plot_title(boosting_package, mean_abs_rel_diff, undershoot_percentage)
        _plot_relative_difference(pred_test, y_test, title=plot_title)

    return mean_abs_rel_diff, undershoot_percentage


def _fit_custom_regressor(boosting_package: str, *args, **kwargs
                          ) -> Union[CatBoostRegressor, XGBRegressor, LGBMRegressor]:
    boosting_package_2_fit_function = {"catboost": _fit_custom_catboost_regressor,
                                       "xgboost": _fit_custom_xgboost_regressor,
                                       "lightgbm": _fit_custom_lightgbm_regressor}
    fit_function = boosting_package_2_fit_function[boosting_package]
    return fit_function(*args, **kwargs)


def _fit_custom_catboost_regressor(X_train: np.ndarray, y_train: np.ndarray, custom_loss_function: Callable,
                                   n_estimators: int, random_seed: int
                                   ) -> CatBoostRegressor:
    custom_objective = CatboostObjective(loss_function=custom_loss_function)
    model = CatBoostRegressor(loss_function=custom_objective, n_estimators=n_estimators, random_seed=random_seed,
                              eval_metric="MAE", allow_writing_files=False)
    model.fit(X_train, y_train, verbose=True)
    return model


def _fit_custom_xgboost_regressor(X_train: np.ndarray, y_train: np.ndarray, custom_loss_function: Callable,
                                  n_estimators: int, random_seed: int
                                  ) -> XGBRegressor:
    custom_objective = XgboostObjective(loss_function=custom_loss_function)
    model = XGBRegressor(objective=custom_objective, n_estimators=n_estimators, random_state=random_seed)
    model.fit(X_train, y_train, eval_metric="mae", eval_set=[(X_train, y_train)], verbose=True)
    return model


def _fit_custom_lightgbm_regressor(X_train: np.ndarray, y_train: np.ndarray, custom_loss_function: Callable,
                                   n_estimators: int, random_seed: int
                                   ) -> LGBMRegressor:
    custom_objective = LightGbmObjective(loss_function=custom_loss_function)
    model = LGBMRegressor(objective=custom_objective, n_estimators=n_estimators, random_state=random_seed)
    model.fit(X_train, y_train, eval_metric="mae", eval_set=(X_train, y_train), verbose=True)
    return model


def _calc_relative_difference(pred_test: np.ndarray, y_test: np.ndarray) -> float:
    rel_diff = (pred_test - y_test) / y_test
    return rel_diff


def _calc_mean_absolute_relative_difference(pred_test: np.ndarray, y_test: np.ndarray) -> float:
    rel_diff = _calc_relative_difference(pred_test, y_test)
    mean_abs_rel_diff = float(np.mean(np.abs(rel_diff)))
    return mean_abs_rel_diff


def _calc_undershoot_percentage(pred_test: np.ndarray, y_test: np.ndarray) -> float:
    rel_diff = _calc_relative_difference(pred_test, y_test)
    undershoot_percentage = float(np.mean(rel_diff < 0))
    return undershoot_percentage


def _plot_relative_difference(pred_test: np.ndarray, y_test: np.ndarray, title: str) -> None:
    rel_diff = _calc_relative_difference(pred_test, y_test)

    plt.hist(rel_diff, bins=np.arange(-1, 1.51, 0.1))
    plt.axvline(0, linestyle='--', color='k')
    plt.xlabel("relative difference = (preds - targets) / targets")
    plt.title(title)
    plt.show()


def _print_title(boosting_package: str) -> None:
    title = f"Training and evaluating custom '{boosting_package}' regressor"
    underline = '=' * len(title)
    print('\n\n' + title + '\n' + underline)


def _generate_plot_title(boosting_package: str, mean_abs_rel_diff: float, undershoot_percentage: float) -> str:
    plot_title = f"{boosting_package} regressor\n" \
                 f"mean absolute relative difference = {mean_abs_rel_diff:.2f}\n" \
                 f"undershoot percentage = {undershoot_percentage:.2f}"
    return plot_title


def _plot_custom_loss(custom_loss_function: Callable) -> None:
    preds = np.linspace(-0.5, 0.5, 101)
    targets = np.zeros_like(preds)
    with torch.no_grad():
        loss = np.hstack([custom_loss_function(Tensor([p]), Tensor([t])).numpy()
                          for p, t in zip(preds, targets)])
    plt.plot(preds, loss)
    plt.title("Custom regression loss function - don't undershoot!")
    plt.xlabel("diff = (preds - targets)")
    plt.ylabel("loss")
    plt.show()


if __name__ == '__main__':
    main()



import numpy as np


class OfficialLoglossObjective:
    """ An official Catboost example """

    def calc_ders_range(self, approxes, targets, weights=None):
        assert len(approxes) == len(targets)
        if weights is not None:
            assert len(weights) == len(approxes)

        result = []
        for index in range(len(targets)):
            e = np.exp(approxes[index])
            p = e / (1 + e)
            der1 = targets[index] - p
            der2 = -p * (1 - p)

            if weights is not None:
                der1 *= weights[index]
                der2 *= weights[index]

            result.append((der1, der2))
        return result


class OfficialMseObjective:
    """ An official Catboost example """

    def calc_ders_range(self, approxes, targets, weights):
        assert len(approxes) == len(targets)
        if weights is not None:
            assert len(weights) == len(approxes)

        result = []
        for index in range(len(targets)):
            der1 = targets[index] - approxes[index]
            der2 = -1

            if weights is not None:
                der1 *= weights[index]
                der2 *= weights[index]

            result.append((der1, der2))
        return result



from typing import Union, Callable

import numpy as np
import torch
from sklearn.datasets import load_breast_cancer
from sklearn.metrics import classification_report
from sklearn.metrics import f1_score
from sklearn.model_selection import train_test_split
from torch import Tensor
from torch.nn import SoftMarginLoss

from catboost import CatBoostClassifier
from lightgbm import LGBMClassifier
from xgboost import XGBClassifier

from treeboost_autograd import CatboostObjective, LightGbmObjective, XgboostObjective


def main():
    train_and_eval_custom_classifier(boosting_package="catboost", custom_loss_function=squared_hinge_loss)
    train_and_eval_custom_classifier(boosting_package="xgboost", custom_loss_function=squared_hinge_loss)
    train_and_eval_custom_classifier(boosting_package="lightgbm", custom_loss_function=squared_hinge_loss)


def soft_margin_loss(preds: Tensor, targets: Tensor) -> Tensor:
    targets = 2 * targets - 1
    loss = SoftMarginLoss(reduction="sum")(preds, targets)
    return loss


def squared_hinge_loss(preds: Tensor, targets: Tensor) -> Tensor:
    targets = 2 * targets - 1
    loss = torch.max(Tensor([0]), 1 - preds * targets) ** 2
    loss = loss.sum()
    return loss


def train_and_eval_custom_classifier(boosting_package: str, custom_loss_function: Callable[[Tensor, Tensor], Tensor],
                                     n_estimators: int = 10, random_seed: int = 2021) -> float:
    assert boosting_package in ["catboost", "xgboost", "lightgbm"]
    _print_title(boosting_package)

    X, y = load_breast_cancer(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=random_seed)

    model = _fit_custom_classifier(boosting_package,
                                   X_train, y_train, custom_loss_function,
                                   n_estimators, random_seed)

    pred_test = model.predict(X_test)
    macro_f1_score = f1_score(y_test, pred_test, average="macro")

    print(classification_report(y_test, pred_test))
    return macro_f1_score


def _fit_custom_classifier(boosting_package: str, *args, **kwargs
                           ) -> Union[CatBoostClassifier, XGBClassifier, LGBMClassifier]:
    boosting_package_2_fit_function = {"catboost": _fit_custom_catboost_classifier,
                                       "xgboost": _fit_custom_xgboost_classifier,
                                       "lightgbm": _fit_custom_lightgbm_classifier}
    fit_function = boosting_package_2_fit_function[boosting_package]
    return fit_function(*args, **kwargs)


def _fit_custom_catboost_classifier(X_train: np.ndarray, y_train: np.ndarray, custom_loss_function: Callable,
                                    n_estimators: int, random_seed: int
                                    ) -> CatBoostClassifier:
    custom_objective = CatboostObjective(loss_function=custom_loss_function)
    model = CatBoostClassifier(loss_function=custom_objective, n_estimators=n_estimators, random_seed=random_seed,
                               eval_metric="ZeroOneLoss", allow_writing_files=False)
    model.fit(X_train, y_train, verbose=True)
    return model


def _fit_custom_xgboost_classifier(X_train: np.ndarray, y_train: np.ndarray, custom_loss_function: Callable,
                                   n_estimators: int, random_seed: int
                                   ) -> XGBClassifier:
    custom_objective = XgboostObjective(loss_function=custom_loss_function)
    model = XGBClassifier(objective=custom_objective, n_estimators=n_estimators, random_state=random_seed)
    model.fit(X_train, y_train, eval_metric="error", eval_set=[(X_train, y_train)], verbose=True)
    return model


def _fit_custom_lightgbm_classifier(X_train: np.ndarray, y_train: np.ndarray, custom_loss_function: Callable,
                                    n_estimators: int, random_seed: int
                                    ) -> LGBMClassifier:
    custom_objective = LightGbmObjective(loss_function=custom_loss_function)
    model = LGBMClassifier(objective=custom_objective, n_estimators=n_estimators, random_state=random_seed)
    model.fit(X_train, y_train, eval_metric="binary_error", eval_set=(X_train, y_train), verbose=True)
    return model


def _print_title(boosting_package: str) -> None:
    title = f"Training and evaluating custom '{boosting_package}' classifier"
    underline = '=' * len(title)
    print('\n\n' + title + '\n' + underline)


if __name__ == '__main__':
    main()
